/**
  *  \file server/router.cc
  *  \brief PCC2 Router
  *
  *  This provides session management/routing. It starts Server
  *  processes, manages their lifetime, and routes requests to them.
  *
  *  This module is Unix-specific.
  */

#include <list>
#include <memory>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <time.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "cpluslib/format.h"
#include "server/session.h"
#include "server/request.h"
#include "server/config.h"
#include "util/misc.h"
#include "server/net.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "9999";
static string_t arg_server = "./c2server";
static int arg_timeout = 10000;
static int arg_virgintimeout = 60;
static int arg_sessionlimit = 10;
static bool arg_newsessionswin = false;

static uint32_t session_counter;
static std::list<RouterSession*> sessions;

static void handleTimeouts();

static RouterSession*
handleNewSessionRequest(string_t arg, const char*& msg)
{
    /* Create new session object */
    std::auto_ptr<RouterSession> n(new RouterSession(arg, ++session_counter));

    /* Check and resolve conflicts */
    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        if (int result = n->checkConflict(**i, arg_newsessionswin)) {
            /* We have a conflict */
            if (result > 0) {
                /* New sessions wins */
                delete *i;
                i = sessions.erase(i);
            } else {
                /* Old sessions wins */
                msg = "453 Conflicting session already active\n";
                return 0;
            }
        } else {
            ++i;
        }
    }

    /* Check for overload */
    if (sessions.size() >= arg_sessionlimit) {
        handleTimeouts();
    }
    if (sessions.size() >= arg_sessionlimit) {
        msg = "451 Too many sessions\n";
        return 0;
    }

    /* Start the session */
    if (n->start(arg_server.c_str())) {
        /* Session started successfully */
        RouterSession* nn = n.release();
        sessions.push_back(nn);
        return nn;
    } else {
        /* Session start failed */
        return 0;
    }
}

static void
handleListRequest(int fd)
{
    char buf[100];
    sprintf(buf, "200 OK, %d sessions\n", sessions.size());
    writeStr(fd, buf);
    time_t now = time(0);
    for (std::list<RouterSession*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
        writeStr(fd,
                 format("%10d  %6d %6ds  %c%c %s\n")
                 << (*i)->getSessionId()
                 << (*i)->getPid()
                 << int32_t(difftime(now, (*i)->getLastAccessTime()))
                 << ((*i)->wasUsed() ? '.' : 'V')
                 << ((*i)->wasModified() ? '.' : 'S')
                 << (*i)->getName());
    }
}

static void
handleTalkRequest(Request& req, int fd, int32_t session)
{
    /* Locate the session */
    for (std::list<RouterSession*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
        if ((*i)->getSessionId() == session) {
            /* Found it. Execute request. */
            RouterSession* s = *i;
            s->executeRequest(req, fd);
            if (!s->isActive()) {
                /* This command caused the session to terminate. An
                   answer for the client has already been produced.
                   So all we need to do is to delete the session. */
                delete *i;
                sessions.erase(i);
            }
            return;
        }
    }

    /* FIXME: should still read complete request */
    writeStr(fd, "452 Session timed out\n");
}

static void
handleCloseRequest(int fd, int32_t session)
{
    /* Close the session */
    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        if ((*i)->getSessionId() == session) {
            delete *i;
            i = sessions.erase(i);
        } else {
            ++i;
        }
    }

    /* Return success */
    writeStr(fd, "200 OK\n");
}

static void
handleClosePatternRequest(int fd, const char* pat)
{
    /* Close the session */
    int n = 0;
    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        if ((*i)->checkConflict(pat, false) < 0) {
            delete *i;
            i = sessions.erase(i);
            ++n;
        } else {
            ++i;
        }
    }

    /* Return success */
    writeStr(fd, format("200 OK, %d sessions closed\n", n));
}

static void
handleRestartRequest(int fd, int32_t session)
{
    /* Find and restart the session */
    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        if ((*i)->getSessionId() == session) {
            if ((*i)->restart(arg_server.c_str())) {
                writeStr(fd, "200 OK\n");
            } else {
                delete *i;
                i = sessions.erase(i);
                writeStr(fd, "500 Restart failed\n");
            }
            return;
        } else {
            ++i;
        }
    }

    /* Session not found means it likely timed out */
    writeStr(fd, "452 Session timed out\n");
}

static void
handleRequest(int fd)
{
    Request req;

    /* Read command */
    char* command;
    while ((command = req.readLine()) == 0) {
        if (req.readMore(fd) != Request::sSuccess) {
            writeStr(fd, "400 Bad request\n");
            return;
        }
    }

    /* Find command verb */
    string_t verb;
    while (*command != 0 && *command != ' ' && *command != '\t') {
        verb += toupper(uint8_t(*command++));
    }
    while (*command == ' ' || *command == '\t') {
        ++command;
    }

    /* Process request */
    if (verb == "LIST") {
        handleListRequest(fd);
    } else if (verb == "S") {
        if (command[strspn(command, "0123456789")] != 0) {
            /* FIXME: should still read complete request */
            writeStr(fd, "400 Bad request\n");
        } else {
            handleTalkRequest(req, fd, atoi(command));
        }
    } else if (verb == "CLOSE") {
        if (command[0] == '-') {
            handleClosePatternRequest(fd, command);
        } else if (command[strspn(command, "0123456789")] != 0) {
            writeStr(fd, "400 Bad request\n");
        } else {
            handleCloseRequest(fd, atoi(command));
        }
    } else if (verb == "RESTART") {
        if (command[strspn(command, "0123456789")] != 0) {
            writeStr(fd, "400 Bad request\n");
        } else {
            handleRestartRequest(fd, atoi(command));
        }
    } else if (verb == "NEW") {
        const char* msg = "403 Forbidden\n";
        RouterSession* s = handleNewSessionRequest(command, msg);
        if (s != 0) {
            writeStr(fd, format("201 %d Created\n", s->getSessionId()));
        } else {
            writeStr(fd, msg);
        }
    } else if (verb == "CONFIG") {
        writeStr(fd, "200 OK\n");
        writeStr(fd, format("Router.Host=%s\n"
                            "Router.Port=%s\n"
                            "Router.Timeout=%d\n"
                            "Router.VirginTimeout=%d\n"
                            "Router.MaxSessions=%d\n"
                            "Router.NewSessionsWin=%d\n")
                 << arg_listen_host
                 << arg_listen_port
                 << arg_timeout
                 << arg_virgintimeout
                 << arg_sessionlimit
                 << arg_newsessionswin);
    } else {
        writeStr(fd, "400 Bad request\n");
    }
}

static void
handleTimeouts()
{
    time_t now = time(0);

    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        int timeout = (*i)->wasUsed() ? arg_timeout : arg_virgintimeout;
        if (difftime(now, (*i)->getLastAccessTime()) > timeout) {
            printf("c2router: session %d timed out\n", int((*i)->getSessionId()));
            delete *i;
            i = sessions.erase(i);
        } else {
            ++i;
        }
    }
}

static void
handleRouterConfig(string_t key, string_t value)
{
    if (key == "ROUTER.HOST") {
        arg_listen_host = value;
    } else if (key == "ROUTER.PORT") {
        arg_listen_port = value;
    } else if (key == "ROUTER.SERVER") {
        arg_server = value;
    } else if (key == "ROUTER.TIMEOUT") {
        stringValue(value, arg_timeout);
    } else if (key == "ROUTER.VIRGINTIMEOUT") {
        stringValue(value, arg_virgintimeout);
    } else if (key == "ROUTER.MAXSESSIONS") {
        stringValue(value, arg_sessionlimit);
    } else if (key == "ROUTER.NEWSESSIONSWIN") {
        arg_newsessionswin = isTrue(value);
    }
}

int main()
{
    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Load configuration */
    loadConfiguration(handleRouterConfig);

    /* Where do we want to listen? */
    int sock = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (sock < 0) {
        fprintf(stderr, "unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("c2router: listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(sock, F_SETFD, FD_CLOEXEC);
    listen(sock, 10);
    while (1) {
        int fd = accept(sock, 0, 0);
        if (fd < 0) {
            perror("accept");
            break;
        }
        fcntl(fd, F_SETFD, FD_CLOEXEC);
        // sleep(1);
        handleRequest(fd);
        close(fd);
        handleTimeouts();
    }
    close(sock);
}
